
// (c) COPYRIGHT URI/MIT 1996
// Please read the full copyright statement in the file COPYRIGH.  
//
// Authors:
//      jhrg,jimg       James Gallagher (jgallagher@gso.uri.edu)

// Implementation for the class ClientStructure. See ClientByte.cc
//
// jhrg 9/26/96

// $Log: ClientSequence.cc,v $
// Revision 1.16  1998/11/20 08:41:22  jimg
// Fixed the fix for the extra newline separators; it was broken for arrays
//
// Revision 1.15  1998/11/19 20:53:33  jimg
// Fixed the extra newline bug.
//
// Revision 1.14  1998/09/16 23:02:21  jimg
// Removed write_val() and replaced it with print_val() and print_all_vals()
//
// Revision 1.13  1998/08/03 16:39:49  jimg
// Fixed write_val mfunc so that outermost is no longer used
//
// Revision 1.12  1998/06/09 04:08:02  jimg
// Fixed a but in write_val where deserialize was not called at the top of
// the loop that reads values.
//
// Revision 1.11  1997/10/04 00:40:05  jimg
// Added changes to support the new deserialize() mfunc and write_val().
//
// Revision 1.10  1997/07/16 00:29:12  jimg
// Fixed Sequences::length return type
//
// Revision 1.9  1997/04/19 00:54:56  jimg
// Changed ouput stream to simplify reading (for loaddods).
//
// Revision 1.8  1997/02/06 20:43:40  jimg
// Fixed log messages
//
// Revision 1.7  1997/01/10 06:49:36  jimg
// Changed call to name_map::lookup() so that non-alphanumerics are mapped to
// underscore. 
//
// Revision 1.6  1996/11/23 05:12:11  jimg
// Added support for variable renaming via the name_map object.
//
// Revision 1.5  1996/10/29 21:43:21  jimg
// Always call smart_newline after each field of a Sequence
//
// Revision 1.4  1996/10/25 23:09:43  jimg
// Fixed count_elements so that it works correctly depending on how FLATTEN is
// defined.
// Fixed write_val.
//
// Revision 1.3  1996/10/23 23:46:43  jimg
// Modified so that writeval outputs a `recursive' data stream. Thus any
// data type can be written and later read without being flattened. This
// change was made so that the NSCAT data which is represented as a Sequence
// of Structures.
//
// Revision 1.2  1996/10/07 21:02:33  jimg
// Fixed Header.
// Replaced code in write_val with static functions (which will work better
// once the software handles Sequences with ctor types).
//
// Revision 1.1  1996/09/30 23:59:15  jimg
// Added.

#ifdef _GNUG_
#pragma implementation
#endif

#include <assert.h>

#include <iostream.h>
#include <Pix.h>
#include <SLList.h>
#include <String.h>

#include "debug.h"
#include "ClientSequence.h"
#include "name_map.h"

extern name_map names;
extern bool translate;

extern void smart_newline(ostream &os, Type type);

Sequence *
NewSequence(const String &n)
{
    return new ClientSequence(n);
}

BaseType *
ClientSequence::ptr_duplicate()
{
    return new ClientSequence(*this);
}

ClientSequence::ClientSequence(const String &n) : Sequence(n)
{
}

ClientSequence::~ClientSequence()
{
}

bool 
ClientSequence::read(const String &, int &)
{
    assert(false);
    return false;
}

int
ClientSequence::length()
{
    return -1;
}

// Count the non-ctor elements in the Sequence recurring on embedded
// ctor types.

static int
count_elements(ClientSequence *sp, bool flat)
{
    int i = 0;

    for (Pix p = sp->first_var(); p; sp->next_var(p))
	switch (sp->var(p)->type()) {
	  case dods_byte_c:
	  case dods_int16_c:
	  case dods_uint16_c:
	  case dods_int32_c:
	  case dods_uint32_c:
	  case dods_float32_c:
	  case dods_float64_c:
	  case dods_str_c:
	  case dods_url_c:
	    i++;
	    break;

	  case dods_array_c:
	  case dods_list_c:
	  case dods_structure_c:
	  case dods_sequence_c:
	  case dods_function_c:
	  case dods_grid_c:
	    if (flat)
		i += count_elements((ClientSequence *)sp->var(p), flat);
	    else
		i++;
	    break;

	  default:
	    return 0;
	}

    return i;
}

void 
ClientSequence::print_val(ostream &os, String space, bool print_decls)
{
    // Print the declarations
    os << type_name() << endl << names.lookup(name(), translate);

    int i = count_elements(this, false);
    os << " " << i << endl;

    for (Pix p = first_var(); p; next_var(p)) {
	var(p)->print_val(os, "", print_decls);
	smart_newline(os, var(p)->type());
    }
}

void
ClientSequence::print_all_vals(ostream &os, XDR *src, DDS *dds, String, bool)
{
    // Print the values. Always print decls with the values for writeval.
    print_val(os, "", true);
    while (deserialize(src, dds)) {
	print_val(os, "", true);
    }
}
