
// (c) COPYRIGHT URI/MIT 1996
// Please read the full copyright statement in the file COPYRIGH.  
//
// Authors:
//      jhrg,jimg       James Gallagher (jgallagher@gso.uri.edu)

// Implementation for ClientStr. See ClientByte.cc
//
// jhrg 9/25/96

// $Log: ClientStr.cc,v $
// Revision 1.19  1998/11/25 01:16:19  jimg
// Now processes -w for warnings. This means that verbose (-v) mode is information and -w is for warnings only
//
// Revision 1.18  1998/11/20 08:41:22  jimg
// Fixed the fix for the extra newline separators; it was broken for arrays
//
// Revision 1.17  1998/11/19 20:53:33  jimg
// Fixed the extra newline bug.
//
// Revision 1.16  1998/09/16 23:02:22  jimg
// Removed write_val() and replaced it with print_val() and print_all_vals()
//
// Revision 1.15  1998/08/03 16:39:50  jimg
// Fixed write_val mfunc so that outermost is no longer used
//
// Revision 1.14  1998/06/09 04:09:13  jimg
// Made some of the diagnostic output about interning NaNs part of the verbose
// option. It will only print when writeval is called with the -v option.
//
// Revision 1.13  1998/02/05 20:14:50  jimg
// DODS now compiles with gcc 2.8.x
//
// Revision 1.12  1997/12/02 20:34:37  jimg
// Fixed/Removed the automatic conversion of strings to numbers. This was a
// bad idea.
//
// Revision 1.11  1997/10/04 00:40:06  jimg
// Added changes to support the new deserialize() mfunc and write_val().
//
// Revision 1.10  1997/06/06 04:02:07  jimg
// Added ASCII output option to writeval. Uses a global `flag'.
//
// Revision 1.9  1997/04/19 00:54:58  jimg
// Changed ouput stream to simplify reading (for loaddods).
//
// Revision 1.8  1997/02/06 19:42:39  jimg
// When converting strings to Float64 values, any string that cannot be
// converted is externalized as NaN.
//
// Revision 1.7  1997/01/13 18:01:50  jimg
// Fixed a bug when the degree part of a DDDMMSSH string has only one or two
// digits.
//
// Revision 1.6  1997/01/10 06:49:37  jimg
// Changed call to name_map::lookup() so that non-alphanumerics are mapped to
// underscore. 
//
// Revision 1.5  1996/11/23 05:12:13  jimg
// Added support for variable renaming via the name_map object.
//
// Revision 1.4  1996/11/13 18:02:54  jimg
// Fixed bad usage of strtod(). When strtod() returns 0.0, the pointer passed
// as a second argument must be compared to the pointer to the string to be
// converted instead of testing only that it is not null. If the pointer is not
// equal to the string argument then the number *has* converted successfully
// (to 0.0). Otherwise, the string did not convert.
//
// Revision 1.3  1996/10/23 23:46:44  jimg
// Modified so that writeval outputs a `recursive' data stream. Thus any
// data type can be written and later read without being flattened. This
// change was made so that the NSCAT data which is represented as a Sequence
// of Structures.
//
// Revision 1.2  1996/10/07 21:00:29  jimg
// Fixed Headers.
//
// Revision 1.1  1996/09/30 23:59:18  jimg
// Added.

#ifdef __GNUG__
#pragma implementation
#endif

#include "config_writeval.h"

#include <assert.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

#include <iostream.h>
#include <Pix.h>
#include <SLList.h>
#include <String.h>

#include "ClientStr.h"
#include "name_map.h"

// We *should* be using <nan.h> to handle generating NaN. However, many
// machines don't have a function that can be used to generate NaN so I've
// written my own (which uses its own bastardized version of the FP union).
// 2/5/97 jhrg

#include "nan_hack.h"

static double
MakeNaN() 
{ 
    dnan X;
    X.nan.exponent = 0x7ff; 
    X.nan.qnan_bit = 0x1; 
    
    return X.d;
}

extern name_map names;
extern bool translate;
extern bool ascii;
extern bool string_to_float;
extern bool dms_to_float;
extern bool verbose;
extern bool warning;

Str *
NewStr(const String &n)
{
    return new ClientStr(n);
}

ClientStr::ClientStr(const String &n) : Str(n)
{
}

BaseType *
ClientStr::ptr_duplicate()
{
    return new ClientStr(*this);
}

bool
ClientStr::read(const String &, int &)
{
    assert(false);
    return false;
}

// Is the string a latitude or longitude in degrees-minutes-seconds form?
// Check for a string of the form DDDMMSSH where DDDMMSS are all digits 0-9
// and H is one of N,S,E,W or the lower case equivalents. If #value# fits
// those criteria, return true, otherwise return false.

static bool
really_dms(String value)
{
    static Regex dms("[0-9][0-9]?[0-9]?[0-5][0-9][0-5][0-9][nsewNSEW]", 1);

    return value.matches(dms);
}

// Given that the string #value# is of the form DDDMMSSH, convert it to
// fractional degrees representation (where MMSS is converted into a fraction
// of a degree). Set the sign so that Latitude North is positive, Latitude
// South is negitive, Longitude East is positive and Longitude West is
// negative.
//
// NB: Assume that the MMSSH part is constant length (i.e., 0 min 0 sec is
// represented 0000) and that N,S,W,E, ... is always present.

static double
convert_dms(String value)
{
    int len = value.length();	// Handle variable length in the degree part.
    String hemisphere = value.at(len-1, 1);
    String seconds = value.at(len-3, 2);
    String minutes = value.at(len-5, 2);
    String degrees = value.at(0, len-5);

#if DODS_DEBUG
    cerr << "Hemi: " << hemisphere << endl
	<< "Sec: " << seconds << endl
	<< "Min: " << minutes << endl
	<< "Deg: " << degrees << endl;
#endif

    double secs = atoi((const char *)seconds) / 60.0;
    double mins = (atoi((const char *)minutes) + secs)/ 60.0;
    double val = (double)atoi((const char *)degrees) + mins;

    hemisphere.downcase();
    if (hemisphere == "s" || hemisphere == "w")
	val *= -1.0;

#if DODS_DEBUG
    cerr << "Val: " << val << " (mins : " << mins << " secs: " << secs
	 << ")" <<endl;
#endif

    return val;
}

void 
ClientStr::print_val(ostream &os, String, bool print_decl_p)
{
    bool print_as_float = false;
    double val = 0.0;

    // Old degree-min-sec translation code, an anachronism. jhrg 1/9/98.
    if (dms_to_float && really_dms(_buf)) {
	val = convert_dms(_buf);
	print_as_float = true;
    }

    // Translate all String variables to Floats. jhrg 1/9/98.
    if (string_to_float) {
	char *ptr = NULL;
	const char *str = (const char *)_buf;
	val = strtod(str, &ptr);
	print_as_float = !(val == 0.0 && (ptr == str));

	if (!print_as_float) {
	    val = MakeNaN();
	    print_as_float = true;
	    if (warning) {
		cerr << "Could not translate `" << _buf << "' to a Float64,"
		     << endl;
		cerr << "interning as NaN (not a number: " << val 
		     << ")" << endl;
	    }
	}
    }

    if (print_as_float) {
	if (print_decl_p)
	    os << "Float64" << endl << names.lookup(name(), translate) << endl;

	if (ascii)
	    os << val << " ";
	else
	    os.write((void *)&val, sizeof(double));
    }
    else {
	if (print_decl_p)
	    os << type_name() << endl << names.lookup(name(), translate) 
		<< endl;

	os << _buf << endl;
    }
}
