
// (c) COPYRIGHT URI/MIT 1996,1997
// Please read the full copyright statement in the file COPYRIGH.  
//
// Authors:
//	jhrg,jimg	James Gallagher (jgallagher@gso.uri.edu)

// writeval is a simple program which reads DODS datasets and writes the
// binary data to stdout using a simple scheme. The program expects one or
// more URLs will be given as arguments. Each URL will be dereferenced in the
// order given on the command line. Data is writen to stdout prefixed by data
// type and size information.

// $Log: write_val.cc,v $
// Revision 1.31  1998/11/25 01:16:19  jimg
// Now processes -w for warnings. This means that verbose (-v) mode is information and -w is for warnings only
//
// Revision 1.30  1998/11/20 08:41:22  jimg
// Fixed the fix for the extra newline separators; it was broken for arrays
//
// Revision 1.29  1998/11/19 20:54:54  jimg
// Fixed the extra newline bug.
//
// Revision 1.28  1998/11/10 03:18:55  jimg
// moved the version number up one notch
//
// Revision 1.27  1998/09/16 23:29:29  jimg
// Fixed the call to ClientStructure::print_all_vals().
//
// Revision 1.26  1998/09/16 23:02:45  jimg
// Removed write_val() and replaced it with print_val() and print_all_vals()
//
// Revision 1.25  1998/08/03 16:39:55  jimg
// Fixed write_val mfunc so that outermost is no longer used
//
// Revision 1.24  1998/06/09 04:10:03  jimg
// Verbose is now a global flag.
//
// Revision 1.23  1998/06/04 06:33:16  jimg
// writeval now prints any message from the Connect's error object if the
// returned dds is null (a flag used to indicate read errors). This means that
// `host not found' errors will be reported to loaddods, etc.
//
// Revision 1.22  1998/03/19 23:27:18  jimg
// Added code to write out error messages read from the server. The format of
// this output is described in the documentation for write_val.
//
// Revision 1.21  1998/02/05 20:14:54  jimg
// DODS now compiles with gcc 2.8.x
//
// Revision 1.20  1997/10/04 00:40:14  jimg
// Added changes to support the new deserialize() mfunc and write_val().
//
// Revision 1.19  1997/09/10 22:51:05  jimg
// Added a `print das' option (-b). This prints the DAS object to stdout.
//
// Revision 1.18  1997/09/05 17:19:07  jimg
// The -A option now accepts the -c option. Use -c to `project' one specific
// attribute container from the DAS. This filters out all the other containers
// so that if a given dataset has an attribute container named `NC_GLOBAL' a
// user can request only the attributes in that container. If the named
// container does not exist, then writeval writes nothing to stdout.
//
// Revision 1.17  1997/06/23 18:34:27  jimg
// Fixed up the online help.
// Added a DDS + data option.
// Updated the version number to 1.4.
//
// Revision 1.16  1997/06/09 20:47:07  jimg
// Fixed a bug in process_per_url_options().
//
// Revision 1.15  1997/06/06 04:06:54  jimg
// Massive changes/additions. Now can read the DODS data stream from stdin
// *and* can write ASCII output (Same output `grammar' as with the binary mode
// but the data is encoded as ASCII). Thus writeval can now serve as a
// server-side filter and a client-side post-retrieval processing agent.
// Updated version to 1.3.
// Added a new option -a to trigger the ASCII output.
//
// Revision 1.14  1997/04/21 20:28:46  jimg
// Updated version number to 1.2 - this matches version 2.13a of DODS.
//
// Revision 1.13  1997/04/19 00:55:01  jimg
// Changed ouput stream to simplify reading (for loaddods).
//
// Revision 1.12  1997/04/08 17:28:30  jimg
// Added test of DDS * returned by Connect::request_data(). This fixes the bug
// where the request results in an error and writeval dumps core after printing
// the message.
//
// Revision 1.11  1997/02/11 16:46:38  jimg
// Removed errant `delete dds' after for loop which processes the variables.
// Don't delete that pointer!
//
// Revision 1.10  1997/02/06 20:46:56  jimg
// Changed verbose to trace and introduced a better verbose mode.
//
// Revision 1.9  1997/02/03 21:25:44  jimg
// Added version information.
//
// Revision 1.8  1997/01/10 06:51:22  jimg
// Added to the online documentation
//
// Revision 1.7  1996/12/18 00:54:20  jimg
// Added code to rename variables and process constraint expressions
// separately from URLs.
//
// Revision 1.6  1996/11/23 05:16:51  jimg
// Added support for per-URL variable renaming. One or more variables can be
// renamed and the set of mappings can be different for each URL.
//
// Revision 1.5  1996/11/13 17:58:40  jimg
// Fixed errors in help text.
//
// Revision 1.4  1996/10/25 23:13:06  jimg
// Added smart_newline() function.
//
// Revision 1.3  1996/10/23 23:49:59  jimg
// Added -f option; map Byte and Int32 types to Float64 and map String to
// Float64 when it can be successfully converted (using strtod()).
// Add changes so that the new write_val(...) (which writes recursive data out)
// will be called.
//
// Revision 1.2  1996/10/07 21:13:51  jimg
// Fixed header.
// Removed redundent handling of embedded constraints (that is not handled by
// the Connect class).
//
// Revision 1.1  1996/09/30 23:59:41  jimg
// Added.

#include "config_dap.h"

static char rcsid[] __unused__ = {"$Id: write_val.cc,v 1.31 1998/11/25 01:16:19 jimg Exp $"};

#include <stdio.h>
#include <assert.h>

#include <GetOpt.h>
#include <Pix.h>
#include <SLList.h>
#include <String.h>

#include "BaseType.h"
#include "Connect.h"

#include "ClientByte.h"
#include "ClientInt32.h"
#include "ClientUInt32.h"
#include "ClientFloat64.h"
#include "ClientStr.h"
#include "ClientUrl.h"
#include "ClientList.h"
#include "ClientArray.h"
#include "ClientStructure.h"
#include "ClientSequence.h"
#include "ClientFunction.h"
#include "ClientGrid.h"
#include "name_map.h"

name_map names;
bool ascii = false;
bool translate = false;
bool string_to_float = false;
bool dms_to_float = false;
bool verbose = false;
bool warning = false;

const char *VERSION = "DODS writeval version: 1.7";

void print_select_attributes(DAS &das, const String &name, ostream &os);
void print_attributes(DAS &das, const String &dataset_name, ostream &os);

static void
usage(String name)
{
    cerr << "Usage: " << name 
	 << " [avfFpdgVdDtw] -- [<url> [-c <expr>] [[-r <var>:<newvar>] ...] ...]" 
	 << endl
	 << name << " [avfgVt] [[-r <var>:<newvar>] ...]" << endl;
    //    cerr << "       a: Use Connect's asynchronous mode." << endl;
    cerr << "       v: Verbose output." << endl;
    cerr << "       a: ASCII output." << endl;
    cerr << "       f: Turn on name canonicalization." << endl; 
    cerr << "       F: Force all string variables to be translated to Float64."
	 << endl;
    cerr << "       p: Interpret Deg-min-sec strings and translate to Float64."
	 << endl
	 << "          This takes precedence over -F if -F is given."
	 << endl;
    cerr << "       g: Use the GUI to show progress" << endl;
    cerr << "       V: Print the version number of this program and exit" 
	 << endl;
    cerr << "       t: Trace network I/O (HTTP, DNS, ...). See geturl"
	 << endl;  
    cerr << "       d: Include type information (DDS object) with the data" 
	 << endl;
    cerr << "       D: Print only the DDS; include no data" << endl;
    cerr << "       A: Return the DAS object" << endl;
    cerr << "       b: Print the DAS object" << endl;
    cerr << "       c: Per-URL constraint expression, enclosed in quotes." 
	 << endl;
    cerr << "       r: Per-URL name mappings; var becomes newvar." << endl;
    cerr << "       w: Show warnings." << endl;
    cerr << endl;
    cerr << "By default writeval dereferences the URL and writes the data" 
	 << endl
	 << "in binary on stdout. Use the -D option to request information"
	 << endl
	 << "about the variable's data types instead, and -A to request" 
	 << endl
	 << "Attribute information." << endl
	 << "Note that -d, -D, -A and -c work only when writeval is given"
	 << endl
	 << "one or more URLs. Use the second form of the command" 
	 << "when reading from standard input." << endl;
}

static String
name_from_url(String url)
{
    // find the last part of the URL (after the last `/') and then strip off
    // the trailing extension (anything following the `.')
    int start = url.index("/", -1) + 1;
    int end = url.index(".", -1);
    
    String name = url.at(start, end-start);

    return name;
}

// Not that smart... Print a new line after writing one of the simple data
// types. This is called by the ctor's write_val() mfuncs so that individual
// values are terminated properly while individual values in an array are not
// terminated.

void
smart_newline(ostream &os, Type type)
{
    switch (type) {
      case dods_byte_c:
      case dods_int16_c:
      case dods_uint16_c:
      case dods_int32_c:
      case dods_uint32_c:
      case dods_float32_c:
      case dods_float64_c:
      case dods_array_c:
      case dods_list_c:
	os << endl;
	break;

	// Strings are lumped in with the ctors because strings end in a
	// new line character and therefore don't need one appended.
      case dods_str_c:
      case dods_url_c:
      case dods_sequence_c:
      case dods_structure_c:
      case dods_function_c:
      case dods_grid_c:
      default:
	break;
    }
    
    os.flush();
}

static void
process_data(Connect &url, DDS *dds)
{

    if (verbose)
	cerr << "Server version: " << url.server_version() << endl;

    for (Pix q = dds->first_var(); q; dds->next_var(q)) {
	switch (dds->var(q)->type()) {
	  case dods_structure_c:
	    (dynamic_cast<ClientStructure *>
	     (dds->var(q)))->print_all_vals(cout, url.source(), dds);
	    break;

	  case dods_sequence_c:
	    (dynamic_cast<ClientSequence *>
	     (dds->var(q)))->deserialize(url.source(), dds); 
	    (dynamic_cast<ClientSequence *>
	     (dds->var(q)))->print_all_vals(cout, url.source(), dds);
	    break;

	  default:
	    dds->var(q)->print_val(cout, "", true);
	    break;
	}

	smart_newline(cout, dds->var(q)->type());
    }
}

static void
process_per_url_options(int &i, int argc, char *argv[])
{
    names.delete_all();	// Clear the global name map for this URL.

    // Test for per-url option. Set variables accordingly.
    while (argv[i+1] && argv[i+1][0] == '-')
	switch (argv[++i][1]) {
	  case 'r':
	    ++i;	// Move past option to argument.
	    if (verbose)
		cerr << "  Renaming: " << argv[i] << endl;
	    // Note that NAMES is a global variable so that all the
	    // writeval() mfuncs can access it without having to pass
	    // it into each function call.
	    names.add(argv[i]);
	    break;
		
	  case 'c':
	    ++i;
	    if (verbose)
		cerr << "  Constraint: " << argv[i] << " ignored." << endl;
	    break;
		
	  default:
	    cerr << "Unknown option `" << argv[i][1] 
		 << "' paired with URL has been ignored." << endl;
	    break;
	}

}

/** Write out the given error object. If the Error object #e# is empty, don't
    write anything out (since that will confuse loaddods).

    @author jhrg */

static void
output_error_object(Error e)
{
    if (e.OK()) {
	cout << "Error" << endl;
	cout << e.error_message() << endl << endl;
    }
}

int
main(int argc, char * argv[])
{
    GetOpt getopt (argc, argv, "afFpdDbAgvVwt:");
    int option_char;
    bool cexpr = false;
    bool trace = false;
    bool gui = false;
    bool dds = false;
    bool das = false;
    bool dds_only = false;
    bool das_only = false;
    bool print_das = false;
    String expr = "";
    char *tcode = NULL;
    int topts = 0;
    // Connect dummy("http");

    putenv("_POSIX_OPTION_ORDER=1"); // Suppress GetOpt's argv[] permutation.

    while ((option_char = getopt()) != EOF)
	switch (option_char) {
	  case 'a': ascii = true; break;
	  case 'v': verbose = true; break;
	  case 'w': warning = true; break;
	  case 't':
	    trace = true;
	    topts = strlen(getopt.optarg);
	    if (topts) {
		tcode = new char[topts + 1];
		strcpy(tcode, getopt.optarg); 
	    }
	    break;
	  case 'f': translate = true; break;
	  case 'F': string_to_float = true; break;
	  case 'p': dms_to_float = true; break;
	  case 'g': gui = true; break;
	  case 'd': dds = true; break;
	  case 'D': dds = dds_only = true; break;
	  case 'b': das = print_das = das_only = true; break;
	  case 'A': das = das_only = true; break;
	  case 'V': {cerr << VERSION << endl; exit(0);}
	  case 'h':
	  case '?':
	  default:
	    usage(argv[0]); exit(1); break;
	}

    char c, *cc = tcode;
    if (verbose && topts > 0)
	while ((c = *cc++))
	    switch (c) {
	      case 'a': WWWTRACE |= SHOW_ANCHOR_TRACE; break;
	      case 'b': WWWTRACE |= SHOW_BIND_TRACE; break;
	      case 'c': WWWTRACE |= SHOW_CACHE_TRACE; break;
	      case 'l': WWWTRACE |= SHOW_SGML_TRACE; break;
	      case 'm': WWWTRACE |= SHOW_MEM_TRACE; break;
	      case 'p': WWWTRACE |= SHOW_PROTOCOL_TRACE; break;
	      case 's': WWWTRACE |= SHOW_STREAM_TRACE; break;
	      case 't': WWWTRACE |= SHOW_THREAD_TRACE; break;
	      case 'u': WWWTRACE |= SHOW_URI_TRACE; break;
	      default:
		cerr << "Unrecognized verbose option: `" << c << "'" << endl;
		break;
	    }
    
    delete tcode;

    Connect *url = 0;

    // If after processing all the command line options there is nothing left
    // (no URL or file) assume that we should read from stdin. This test
    // allows `-r' options to be used when reading from a pipe or
    // redirection. It is assumed that all command options will be rmoved by
    // the time execution gets here. Thus, if a optio appears here we must be
    // reading from stdin.
    if (getopt.optind == argc || argv[getopt.optind][0] == '-') {
	if (verbose)
	    cerr << "Assuming standard input is a DODS data stream." << endl;

	url = new Connect("stdin", trace);

	process_per_url_options(getopt.optind, argc, argv);

	DDS *dds = url->read_data(stdin, gui, false);
	if (dds)
	    process_data(*url, dds);
	else
	    output_error_object(url->error());
    }

    for (int i = getopt.optind; i < argc; ++i) {
	if (url)
	    delete url;

	url = new Connect(argv[i]);

	DBG2(cerr << "argv[" << i << "] (of " << argc << "): " << argv[i] \
	     << endl);

	if (url->is_local()) {
	    if (verbose) 
		cerr << "Assuming that the argument " << argv[i] 
		     << " is a file" << endl 
		     << "that contains a DODS data object; decoding." << endl;

	    process_per_url_options(i, argc, argv);

	    DDS *dds = url->read_data(fopen(argv[i], "r"), gui, false);
	    if (dds)
		process_data(*url, dds);
	    else 
		output_error_object(url->error());

	    continue;
	}

	if (verbose)
	    cerr << endl << "Reading: " << url->URL(false) << endl;
	
	// If the user supplied -d or -D, print the DDS
	if (dds) {
	    url->request_dds(gui);
	    url->dds().print();
	}

	// If the user supplied -A, print the DAS
	if (das && !print_das) {
	    // Look for a -c option following the URL; it gives the name of
	    // an attribute container to print. This is a primitive
	    // constraint mechanism that lets the (somewhat) brain-dead
	    // loaddods program ask for a particular container rather than
	    // get all of the containers. Note the gratuitous code
	    // replication... 
	    int j = i;
	    bool found_ce = false;
	    String expr;
	    while (!found_ce && j < argc && argv[j+1] && argv[j+1][0] == '-')
		if (argv[++j][1] == 'c') {
		    found_ce = true;
		    expr = argv[++j];
		    i = j;
		}
	    
	    url->request_das(gui);

	    if (found_ce)
		print_select_attributes(url->das(), expr, cout);
	    else
		print_attributes(url->das(), name_from_url(url->URL()), cout);
	}

	if (das && print_das) {
	    url->request_das(gui);

	    url->das().print();
	}

	// If the users wants only the DDS or DAS, don't do this part (which
	// gets data). However, writeval by default does get data.
	if (!(dds_only || das_only)) {
	    // Scan ahead for a -c option. If it does not exist, print the CE
	    // here before printing the stuff about renaming variables.
	    // Otherwise delay printing the constraint until the in the
	    // second while-loop.
	    int j = i;
	    bool found_ce = false;
	    while (!found_ce && j < argc && argv[j+1] && argv[j+1][0] == '-')
		if (argv[++j][1] == 'c') {
		    found_ce = true;
		    expr = argv[++j];
		    i = j;
		}

	    if (verbose && !found_ce)
		cerr << "  Constraint: " << url->CE() << endl;

	    process_per_url_options(i, argc, argv);

	    DDS *dds = url->request_data(expr, gui, false);

    	    DBG(cerr << "Starting writing data..." << endl);

	    if (dds)
		process_data(*url, dds);
	    else
		output_error_object(url->error());

	    DBG(cerr << "Completed writing data." << endl);
	}

	cout.flush();
    }

    cerr.flush();

    return 0;
}
